from keap.REST.V1.mixins import CreateMixin, DeleteMixin, ListMixin, RetrieveMixin, UpdateMixin


class FileService(CreateMixin, DeleteMixin, ListMixin, RetrieveMixin, UpdateMixin):
    api_url = "appointments"

    def __init__(self, keap):
        super().__init__(keap)



from keap.REST import BaseService
from keap.REST.models import CustomFieldType


class BaseServiceMixin(BaseService):
    primary_key = "id"
    update_verb = "PUT"


class ModelMixin(BaseServiceMixin):
    def model(self):
        return self._get(f"{self.api_url}/model")


class CreateMixin(BaseServiceMixin):
    def create(self, **kwargs):
        return self._post(f"{self.api_url}", data=kwargs)


class DeleteMixin(BaseServiceMixin):
    def delete(self, id: int, params: dict = None):
        return self._delete(f"{self.api_url}/{id}", params=params)


class RetrieveMixin(BaseServiceMixin):
    def get(self, id: int, **kwargs):
        return self._get(f"{self.api_url}/{id}", params=kwargs)


class SearchMixin(BaseServiceMixin):
    def search(self, **kwargs):
        return self._get(f"{self.api_url}/search", params=kwargs)


class UpdateMixin(BaseServiceMixin):
    def update(self, id: int, data: dict = None, params: dict = None):
        return self._put(f"{self.api_url}/{id}", data=data, params=params)


class ReplaceMixin(BaseServiceMixin):
    def replace(self, id: int, data: dict = None, params: dict = None):
        return self._patch(f"{self.api_url}/{id}", data=data, params=params)


class ListMixin(BaseServiceMixin):
    def list(self, **kwargs):
        return self._get(f"{self.api_url}", params=kwargs)


class CreateCustomFieldMixin(BaseServiceMixin):
    def create_custom_field(self,
                            field_type: CustomFieldType,
                            label: str,
                            options: list = None,
                            group_id: int = 0,
                            user_group_id: int = None
                            ):
        data = locals()
        data['field_type'] = field_type.value
        return self._post(f"{self.api_url}/model/customFields", data=data)



from typing import List

from datetime import datetime

from keap.REST.V1.mixins import ModelMixin, DeleteMixin, RetrieveMixin, CreateCustomFieldMixin
from keap.REST.models import Address, CustomField, PhoneNumber, SocialAccount, EmailAddress, DuplicateOption


class ContactService(ModelMixin, DeleteMixin, RetrieveMixin, CreateCustomFieldMixin):
    api_url = "contacts"
    return_key = "contacts"
    update_verb = "patch"

    def __init__(self, keap):
        super().__init__(keap)

    def create(self,
               addresses: List[Address] = None,
               anniversary: datetime = None,
               birthday: datetime = None,
               company: int = None,
               contact_type: str = None,
               custom_fields: List[CustomField] = None,
               duplicate_option: DuplicateOption = None,
               email_addresses: List[EmailAddress] = None,
               family_name: str = None,
               fax_numbers: list = None,
               given_name: str = None,
               lead_source_id: int = None,
               middle_name: str = None,
               opt_in_reason: str = None,
               origin: dict = None,
               owner_id: int = None,
               phone_numbers: List[PhoneNumber] = None,
               preferred_locale: str = None,
               preferred_name: str = None,
               prefix: str = None,
               social_accounts: List[SocialAccount] = None,
               source_type: str = None,
               spouse_name: str = None,
               suffix: str = None,
               time_zone: str = None,
               website: str = None,
               **kwargs):
        data = locals() | kwargs
        if data.get(self.primary_key) or data.get('duplicate_option'):
            return self._put(f"{self.api_url}", data=data)
        return self._post(f"{self.api_url}", data=data)

    def list(self,
             limit: int = 1000,
             offset: int = 0,
             order: str = "id",
             ascending: bool = False,
             email: str = None,
             family_name: str = None,
             given_name: str = None,
             since: datetime = None,
             until: datetime = None,
             optional_properties: list = None
             ):
        params = locals()
        return self._get(f"{self.api_url}", params=params)

    def update(self,
               id: int,
               update_mask: list = None,
               addresses: List[Address] = None,
               anniversary: datetime = None,
               birthday: datetime = None,
               company: int = None,
               contact_type: str = None,
               custom_fields: List[CustomField] = None,
               email_addresses: List[EmailAddress] = None,
               family_name: str = None,
               fax_numbers: list = None,
               given_name: str = None,
               lead_source_id: int = None,
               middle_name: str = None,
               opt_in_reason: str = None,
               origin: dict = None,
               owner_id: int = None,
               phone_numbers: List[PhoneNumber] = None,
               preferred_locale: str = None,
               preferred_name: str = None,
               prefix: str = None,
               social_accounts: List[SocialAccount] = None,
               source_type: str = None,
               spouse_name: str = None,
               suffix: str = None,
               time_zone: str = None,
               website: str = None,
               **kwargs
               ):
        data = locals() | kwargs
        del data['update_mask']

        params = {}
        if update_mask:
            params['update_mask'] = update_mask

        return self._patch(f"{self.api_url}/{id}", data=data, params=params)

    def get_credit_cards(self, id: int):
        return self._get(f"{self.api_url}/{id}/creditCards")

    def create_credit_card(self,
                           id: int,
                           address: Address,
                           card_number: str,
                           card_type: str,
                           email_address: str,
                           expiration_month: str,
                           expiration_year: str,
                           maestro_issue_number: str,
                           maestro_start_date_month: str,
                           maestro_start_date_year: str,
                           name_on_card: str,
                           verification_code: str):
        data = locals()
        return self._post(f"{self.api_url}/{id}/creditCards", data=data)

    def list_emails(self, id: int, email: str = None, limit: int = 1000, offset: int = 0):
        params = {
            'limit': limit,
            'offset': offset,
        }
        if email:
            params['email'] = email
        return self._get(f"{self.api_url}/{id}/emails", params=params)

    def get_tags(self, id: int, limit: int = 1000, offset: int = 0):
        params = {
            'limit': limit,
            'offset': offset,
        }
        return self._get(f"{self.api_url}/{id}/tags", params=params)

    def add_tags(self, id: int, tags: list):
        data = {
            'tagIds': tags,
        }
        return self._post(f"{self.api_url}/{id}/tags", data=data)

    def remove_tags(self, id: int, tags: list):
        params = {
            'ids': tags,
        }
        return self._delete(f"{self.api_url}/{id}/tags", params=params)

    def remove_tag(self, id: int, tag: int):
        return self._delete(f"{self.api_url}/{id}/tags/{tag}")



import json
from enum import Enum
from typing import List, Literal


class BaseDataModel(object):
    @property
    def data(self):
        return self.__dict__


class CustomField(BaseDataModel):
    content: any = None
    id: int = None

    def __init__(self, id, content):
        self.content = content
        self.id = id


class EmailAddress(BaseDataModel):
    field: str
    email: str
    slot_min = 1
    slot_max = 3

    def __init__(self, slot: int):
        if slot < self.slot_min or slot > self.slot_max:
            raise ValueError
        self.field = f"EMAIL{slot}"


class PhoneFieldBase(BaseDataModel):
    field: str = None
    number: str = None
    type: str = None
    extension: str = None
    slot_min = 1
    slot_max = 5

    def __init__(self, slot: int):
        if slot < self.slot_min or slot > self.slot_max:
            raise ValueError
        self.field = f"{self.field}{slot}"


class FaxNumber(PhoneFieldBase):
    slot_min = 1
    slot_max = 2

    def __init__(self, slot: int):
        self.field = "FAX"
        super().__init__(slot)


class PhoneNumber(PhoneFieldBase):
    slot_min = 1
    slot_max = 5

    def __init__(self, slot: int):
        self.field = f"PHONE"
        super().__init__(slot)


class SocialAccount(BaseDataModel):
    name: str
    type: str

    def __init__(self,
                 platform: Literal['Facebook', 'Twitter', 'LinkedIn', 'Instagram', 'Snapchat', 'YouTube', 'Pinterest']):
        if platform not in ['Facebook', 'Twitter', 'LinkedIn', 'Instagram', 'Snapchat', 'YouTube', 'Pinterest']:
            raise ValueError
        self.type = platform


class Address(BaseDataModel):
    country_code: str = ""
    field: str = ""
    line1: str = ""
    line2: str = ""
    locality: str = ""
    postal_code: str = ""
    region: str = ""
    zip_code: str = ""
    zip_four: str = ""

    def __str__(self):
        return f"{self.line1} {self.locality}, {self.region} {self.zip_code}"

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__)


class BillingAddress(Address):
    def __init__(self):
        self.field = "BILLING"


class ShippingAddress(Address):
    def __init__(self):
        self.field = "SHIPPING"


class OtherAddress(Address):
    def __init__(self):
        self.field = "OTHER"


class DuplicateOption(Enum):
    EMAIL = "Email"
    EMAIL_AND_NAME = "EmailAndName"


class CustomFieldType(Enum):
    CURRENCY = 'Currency'
    DATE = 'Date'
    DATETIME = 'DateTime'
    DAY_OF_WEEK = 'DayOfWeek'
    DRILLDOWN = 'Drilldown'
    EMAIL = 'Email'
    MONTH = 'Month'
    LIST_BOX = 'ListBox'
    NAME = 'Name'
    WHOLE_NUMBER = 'WholeNumber'
    DECIMAL_NUMBER = 'DecimalNumber'
    PERCENT = 'Percent'
    PHONE_NUMBER = 'PhoneNumber'
    RADIO = 'Radio'
    DROPDOWN = 'Dropdown'
    SOCIAL_SECURITYN_UMBER = 'SocialSecurityNumber'
    STATE = 'State'
    TEXT = 'Text'
    TEXT_AREA = 'TextArea'
    USER = 'User'
    USER_LIST_BOX = 'UserListBox'
    WEBSITE = 'Website'
    YEAR = 'Year'
    YES_NO = 'YesNo'



from keap.REST.V1.mixins import BaseServiceMixin


class LocaleService(BaseServiceMixin):
    api_url = "locales"

    def __init__(self, keap):
        super().__init__(keap)

    def list_countries(self):
        return self._get(f"{self.api_url}/countries")

    def list_country_provinces(self, country_code: str):
        return self._get(f"{self.api_url}/countries/{country_code}/provinces")

    def list_dropdown_defaults(self):
        return self._get(f"{self.api_url}/defaultOptions")



from keap.XML import DataService, ContactService, AffiliateService, AffiliateProgramService, DiscountService, \
    EmailService, FileService, FunnelService, InvoiceService, OrderService, ProductService, SearchService, \
    ShippingService, WebFormService


class XML:
    def __init__(self, keap):
        self.keap = keap
        self.AffiliateProgramService = AffiliateProgramService(keap)
        self.AffiliateService = AffiliateService(keap)
        self.ContactService = ContactService(keap)
        self.DataService = DataService(keap)
        self.DiscountService = DiscountService(keap)
        self.EmailService = EmailService(keap)
        self.FileService = FileService(keap)
        self.FunnelService = FunnelService(keap)
        self.InvoiceService = InvoiceService(keap)
        self.OrderService = OrderService(keap)
        self.ProductService = ProductService(keap)
        self.SearchService = SearchService(keap)
        self.ShippingService = ShippingService(keap)
        self.WebFormService = WebFormService(keap)

    def test_connection(self):
        return self.DataService.get_app_setting('Application', 'enabled')



from keap.REST.V1.mixins import (CreateMixin, CreateCustomFieldMixin, DeleteMixin, ListMixin, ModelMixin, ReplaceMixin,
                                 RetrieveMixin, SearchMixin, UpdateMixin)


class TaskService(CreateMixin, CreateCustomFieldMixin, DeleteMixin, ListMixin, ModelMixin, ReplaceMixin,
                  RetrieveMixin, SearchMixin, UpdateMixin):
    api_url = "tasks"

    def __init__(self, keap):
        super().__init__(keap)



from keap.REST.V1.mixins import ListMixin, RetrieveMixin


class CampaignService(ListMixin, RetrieveMixin):
    api_url = "campaigns"

    def __init__(self, keap):
        super().__init__(keap)

    def achieve_goal(self, integration: str, call: str, contact_id: int):
        return self._post(f"{self.api_url}/goals/{integration}/{call}", data={'contact_id': contact_id})

    def add_contact_to_campaign(self, campaign_id: int, sequence_id: int, contact_id: int):
        return self._post(f"{self.api_url}/{campaign_id}/sequences/{sequence_id}/contacts/{contact_id}")

    def add_contacts_to_campaign(self, campaign_id: int, sequence_id: int, contacts: list):
        return self._post(f"{self.api_url}/{campaign_id}/sequences/{sequence_id}/contacts",
                          data={'ids': contacts})

    def remove_contact_from_campaign(self, campaign_id: int, sequence_id: int, contact_id: int):
        return self._delete(f"{self.api_url}/{campaign_id}/sequences/{sequence_id}/contacts/{contact_id}")

    def remove_contacts_from_campaign(self, campaign_id: int, sequence_id: int, contacts: list):
        return self._delete(f"{self.api_url}/{campaign_id}/sequences/{sequence_id}/contacts",
                            data={'ids': contacts})



from keap import Keap
from keap.exceptions import KeapUnauthorizedException, KeapException, KeapTokenExpiredException

try:
    from xmlrpclib import ServerProxy, Error
except ImportError:
    from xmlrpc.client import ServerProxy, Error


class BaseService:
    xmlrpc_url = 'https://api.infusionsoft.com/crm/xmlrpc/v1'
    _service = None
    client = None

    def __init__(self, keap: Keap):
        self.keap = keap
        self.get_xmlrpc_client()

    def get_xmlrpc_client(self):
        if not self.keap.token.access_token:
            raise Exception(f"No token set for client {self.keap.app_name}")
        uri = f"{self.xmlrpc_url}?access_token={self.keap.token.access_token}"
        self.client = ServerProxy(uri, use_datetime=self.keap.api_settings.USE_DATETIME,
                                  allow_none=self.keap.api_settings.ALLOW_NONE)
        self.client.error = Error
        return self.client

    def __getattr__(self, method):
        def function(*args):
            return self.call(method, *args)

        return function

    @property
    def service(self):
        return self._service if self._service else self.__class__.__name__

    def call(self, method, *args):
        call = getattr(self.client, f"{self.service}.{method}")
        try:
            return call(self.keap.token.access_token, *args)
        except self.client.error as e:
            error_code = e.errcode
            error_message = e.errmsg
            headers = e.headers
            if error_code == 401:  # Unauthorized
                raise KeapUnauthorizedException(error_message)
            raise KeapException(error_message)

    def server(self):
        return self.client



from keap.XML import BaseService


class DataService(BaseService):
    def __init__(self, keap):
        super().__init__(keap)

    def create(self, table, values):
        return self.call("create", table, values)

    def load(self, table, record, fields):
        return self.call("load", table, record, fields)

    def update(self, table, id, fields):
        return self.call("update", table, id, fields)

    def delete(self, table, id):
        return self.call("delete", table, id)

    def count(self, table, query):
        return self.call("count", table, query)

    def query(self, table, limit, page, query, return_fields, sort_by, ascending):
        return self.call("query", table, limit, page, query, return_fields, sort_by, ascending)

    def find_by_field(self, table, limit, page, field_name, field_value, return_fields):
        return self.call("findByField", table, limit, page, field_name, field_value, return_fields)

    def add_custom_field(self, table, display, type, header_id):
        return self.call("addCustomField", table, display, type, header_id)

    def update_custom_field(self, id, values):
        return self.call("updateCustomField", id, values)

    def get_app_setting(self, module, setting):
        return self.call("getAppSetting", module, setting)



from keap.REST.V1.mixins import (CreateMixin, CreateCustomFieldMixin, DeleteMixin, ListMixin, ModelMixin, ReplaceMixin,
                                 RetrieveMixin, UpdateMixin)


class AppointmentService(CreateMixin, CreateCustomFieldMixin, DeleteMixin, ListMixin, ModelMixin, ReplaceMixin,
                         RetrieveMixin, UpdateMixin):
    api_url = "appointments"

    def __init__(self, keap):
        super().__init__(keap)



from keap.REST.V1.mixins import RetrieveMixin, UpdateMixin
import requests


class HookService(RetrieveMixin, UpdateMixin):
    api_url = "hooks"

    def __init__(self, keap):
        super().__init__(keap)

    def create(self,
               eventKey: str = None,
               hookUrl: str = None,
               ):
        data = locals()
        return self._post(f"{self.api_url}", data=data)
        # return self._post(f"{self.api_url}", data=data)

    def list(self):
        return self._get(f"{self.api_url}")

    def delete(self,
               key: str = None):
        return self._delete(f"{self.api_url}/{key}")

    def verify(self,
               key: int = None):
        return self._post(f'{self.api_url}/{key}/verify')

    def list_event_keys(self):
        return self._get(f"{self.api_url}/event_keys")

    def delete_hooks(self,
                     hook_keys: list = None):
        for key in hook_keys:
            self._delete(key=key)





from keap.REST.V1.mixins import (CreateMixin, DeleteMixin, ListMixin, ModelMixin, RetrieveMixin)


class OrderService(CreateMixin, DeleteMixin, ListMixin, ModelMixin, RetrieveMixin):
    api_url = "orders"

    def __init__(self, keap):
        super().__init__(keap)

    def create_order_item(self, order_id: int, **kwargs):
        return self._get(f"{self.api_url}/{order_id}/items", params=kwargs)

    def delete_order_item(self, order_id: int, order_item_id: int):
        return self._delete(f"{self.api_url}/{order_id}/items/{order_item_id}")

    def replace_order_payplan(self, order_id: int, **kwargs):
        return self._delete(f"{self.api_url}/{order_id}/paymentPlan", data=kwargs)

    def get_order_payments(self, order_id: int):
        return self._get(f"{self.api_url}/{order_id}/payments")

    def create_order_payments(self, order_id: int, **kwargs):
        return self._post(f"{self.api_url}/{order_id}/payments", data=kwargs)

    def get_order_transactions(self, order_id: int, **kwargs):
        return self._get(f"{self.api_url}/{order_id}/transactions", params=kwargs)



import base64
import json
from pathlib import Path
from urllib.parse import urlencode
import requests
from keap.KeapToken import KeapToken
from keap.REST.REST_V1 import REST_V1
from keap.XML.XML import XML
from keap.settings import APISettings
from keap.storages import BaseStorage, JSONStorage


class Keap:
    rest_url = 'https://api.infusionsoft.com/crm/rest/v1/'
    authorize_url = 'https://signin.infusionsoft.com/app/oauth/authorize'
    access_token_url = 'https://api.infusionsoft.com/token'
    app_name: str = None
    storage: BaseStorage = None
    api_settings: APISettings
    xml_client = None
    rest_v1 = None

    def __init__(self, config: dict = None, config_file: Path = None):
        if config and config_file:
            raise ValueError("You can only load settings from one source")
        if config_file:
            with open(config_file, 'r') as f:
                config = json.load(f)
        self.api_settings = APISettings(config)
        if not self.api_settings.CLIENT_ID:
            raise Exception("Missing Client Id")
        if not self.api_settings.CLIENT_SECRET:
            raise Exception("Missing Client Secret")

        self.app_name = self.api_settings.APP_NAME

        self.storage = self.api_settings.STORAGE_CLASS()
        if isinstance(self.api_settings.STORAGE_CLASS(), JSONStorage):
            if not self.api_settings.JSON_STORAGE_PATH:
                raise Exception("JSON_STORAGE_PATH must be defined for JSONStorage")
            self.storage.storage_path = self.api_settings.JSON_STORAGE_PATH

        # self.XML = XML(self)

    @property
    def XML(self):
        if self.token.is_expired:
            self.refresh_access_token()
        if not self.xml_client:
            self.xml_client = XML(self)
        return self.xml_client


    @property
    def REST_V1(self):
        if not self.rest_v1:
            self.rest_v1 = REST_V1(self)
        return self.rest_v1

    @property
    def token(self) -> KeapToken:
        return self.storage.get_token(self.app_name)

    def save_token(self, token):
        self.storage.save_token(self.app_name, token)
        self.xml_client = XML(self)

    def get_authorization_url(self, state=''):
        data = {
            'client_id': self.api_settings.CLIENT_ID,
            'redirect_uri': self.api_settings.REDIRECT_URL,
            'response_type': 'code',
            'scope': 'full',
            'state': state
        }
        return self.authorize_url + "?" + urlencode(data)

    def request_access_token(self, code):
        data = {
            'client_id': self.api_settings.CLIENT_ID,
            'client_secret': self.api_settings.CLIENT_SECRET,
            'redirect_uri': self.api_settings.REDIRECT_URL,
            'code': code,
            'grant_type': 'authorization_code',
        }
        authorization_token_response = requests.post(self.access_token_url, data)
        token = KeapToken(**authorization_token_response.json())
        self.save_token(token)
        return self.token

    def refresh_access_token(self):
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': self.token.refresh_token
        }
        auth_header = "Basic " + base64.b64encode(
            bytes(f"{self.api_settings.CLIENT_ID}:{self.api_settings.CLIENT_SECRET}", 'utf-8')
        ).decode()
        refresh_token_response = requests.post(self.access_token_url, data, headers={'Authorization': auth_header})
        # TODO: Make sure valid token is returned before setting and saving.
        token = KeapToken(**refresh_token_response.json())
        self.save_token(token)
        return self.token

    def change_app(self, app_name):
        self.app_name = app_name
        if not self.token.access_token:
            raise Exception(f"{app_name} does not have a token in storage, please authenticate")
        self.xml_client = XML(self)
        self.rest_v1 = REST_V1(self)



from keap.REST.V1 import (
    AccountService, AffiliateService, AppointmentService, CampaignService, CompanyService,
    ContactService, EmailService, FileService, HookService, LocaleService, MerchantService, NoteService, OpportunityService,
    OrderService, ProductService, SettingService, SubscriptionService, TagService, TaskService, TransactionService,
    UserService)


class REST_V1:
    def __init__(self, keap):
        self.keap = keap
        self.AccountService = AccountService(keap)
        self.AffiliateService = AffiliateService(keap)
        self.AppointmentService = AppointmentService(keap)
        self.CampaignService = CampaignService(keap)
        self.CompanyService = CompanyService(keap)
        self.ContactService = ContactService(keap)
        self.EmailService = EmailService(keap)
        self.FileService = FileService(keap)
        self.HookService = HookService(keap)
        self.LocaleService = LocaleService(keap)
        self.MerchantService = MerchantService(keap)
        self.NoteService = NoteService(keap)
        self.OpportunityService = OpportunityService(keap)
        self.OrderService = OrderService(keap)
        self.ProductService = ProductService(keap)
        self.SettingService = SettingService(keap)
        self.SubscriptionService = SubscriptionService(keap)
        self.TagService = TagService(keap)
        self.TaskService = TaskService(keap)
        self.TransactionService = TransactionService(keap)
        self.UserService = UserService(keap)

    def test_connection(self):
        return self.SettingService.get_application_status()



from keap.REST.V1.mixins import CreateMixin, DeleteMixin, ListMixin, RetrieveMixin


class EmailService(CreateMixin, DeleteMixin, ListMixin, RetrieveMixin):
    api_url = "emails"

    def __init__(self, keap):
        super().__init__(keap)

    def create_email(self, **kwargs):
        self._post(f"{self.api_url}", data=kwargs)

    def list_emails(self, contact_id: int = None, email: str = None, limit: int = None, offset: int = None, ordered: bool = None, since_sent_date: str = None, until_sent_date: str = None):
        params = locals()
        return self._get(f"{self.api_url}", params=params)

    def send_email(self, **kwargs):
        self._post(f"{self.api_url}", data=kwargs)

    def sync_email_records(self, **kwargs):
        self._post(f"{self.api_url}/sync", data=kwargs)

    def unsync_email_records(self, **kwargs):
        self._post(f"{self.api_url}/unsync", data=kwargs)

    def update_email_address(self, email: str, opted_in: bool = True, reason: str = "Opted in via API"):
        self._put(f"emailAddresses/{email}", data={'opted_in': opted_in, 'reason': reason})

    def list_emails_by_contact_id(self, contact_id: int):
        params = locals()
        return self._get(f"{self.api_url}", params=params)

    def get_email(self, id:int):
        return self._get(f"{self.api_url}/{id}")



from keap.REST.V1.mixins import RetrieveMixin, UpdateMixin


class AccountService(RetrieveMixin, UpdateMixin):
    api_url = "account/profile"

    def __init__(self, keap):
        super().__init__(keap)



from keap.XML import BaseService


class EmailService(BaseService):
    _service = "APIEmailService"

    def __init__(self, keap):
        super().__init__(keap)

    def get_optin_status(self, email):
        return self.call("getOptStatus", email)

    def optout_email(self, email, reason):
        return self.call("optOut", email, reason)

    def optin_email(self, email, reason):
        return self.call("optIn", email, reason)



class KeapException(Exception):
    pass


class KeapForbiddenException(KeapException):
    pass


class KeapUnauthorizedException(KeapException):
    pass


class KeapTokenExpiredException(KeapException):
    pass



from .BaseService import BaseService
from .AffiliateProgramService import AffiliateProgramService
from .AffiliateService import AffiliateService
from .ContactService import ContactService
from .DataService import DataService
from .DiscountService import DiscountService
from .EmailService import EmailService
from .FileService import FileService
from .FunnelService import FunnelService
from .InvoiceService import InvoiceService
from .OrderService import OrderService
from .ProductService import ProductService
from .SearchService import SearchService
from .ShippingService import ShippingService
from .WebFormService import WebFormService



from keap.REST.V1.mixins import (CreateMixin, CreateCustomFieldMixin, DeleteMixin, ListMixin, ModelMixin, ReplaceMixin,
                                 RetrieveMixin, UpdateMixin)


class ProductService(CreateMixin, DeleteMixin, ListMixin, ReplaceMixin,
                     RetrieveMixin, UpdateMixin):
    api_url = "products"

    def __init__(self, keap):
        super().__init__(keap)

    def upload_product_image(self, product_id: int, **kwargs):
        return self._post(f"{self.api_url}/{product_id}/image", data=kwargs)

    def delete_product_image(self, product_id: int):
        return self._delete(f"{self.api_url}/{product_id}/image")

    def create_product_subscription(self, product_id: int, **kwargs):
        return self._post(f"{self.api_url}/{product_id}/subscriptions", data=kwargs)

    def get_product_subscription(self, product_id: int, subscription_id: int):
        return self._get(f"{self.api_url}/{product_id}/subscriptions/{subscription_id}")

    def delete_product_subscription(self, product_id: int, subscription_id: int):
        return self._get(f"{self.api_url}/{product_id}/subscriptions/{subscription_id}")



from keap.REST.V1.mixins import ListMixin, CreateMixin, ModelMixin, RetrieveMixin


class AffiliateService(CreateMixin, ListMixin, ModelMixin, RetrieveMixin):
    api_url = "affiliates"

    def __init__(self, keap):
        super().__init__(keap)

    def get_commissions(self, **kwargs):
        return self._get(f"{self.api_url}/commissions", params=kwargs)

    def get_programs(self, **kwargs):
        return self._get(f"{self.api_url}/programs", params=kwargs)

    def get_redirects(self, **kwargs):
        return self._get(f"{self.api_url}/redirectlinks", params=kwargs)

    def get_summaries(self, **kwargs):
        return self._get(f"{self.api_url}/summaries", params=kwargs)

    def get_affilliate_clawbacks(self, id: int, **kwargs):
        return self._get(f"{self.api_url}/{id}/clawbacks", params=kwargs)

    def get_affilliate_payments(self, id: int, **kwargs):
        return self._get(f"{self.api_url}/{id}/payments", params=kwargs)



from keap.REST.V1.mixins import BaseServiceMixin


class SettingService(BaseServiceMixin):
    api_url = "setting"

    def __init__(self, keap):
        super().__init__(keap)

    def get_application_configuration(self):
        return self._get(f"{self.api_url}/application/configuration")

    def get_application_status(self):
        return self._get(f"{self.api_url}/application/enabled")



from keap.XML import BaseService


class AffiliateService(BaseService):
    _service = "APIAffiliateService"

    def __init__(self, keap):
        super().__init__(keap)

    def retrieve_running_totals(self, affiliate_ids):
        return self.affRunningTotals(affiliate_ids)



from .AccountService import AccountService
from .AffiliateService import AffiliateService
from .AppointmentService import AppointmentService
from .CampaignService import CampaignService
from .CompanyService import CompanyService
from .ContactService import ContactService
from .EmailService import EmailService
from .FileService import FileService
from .HookService import HookService
from .LocaleService import LocaleService
from .MerchantService import MerchantService
from .NoteService import NoteService
from .OpportunityService import OpportunityService
from .OrderService import OrderService
from .ProductService import ProductService
from .SettingService import SettingService
from .SubscriptionService import SubscriptionService
from .TagService import TagService
from .TaskService import TaskService
from .TransactionService import TransactionService
from .UserService import UserService



from keap.REST.V1.mixins import (CreateMixin, ListMixin, ModelMixin, ReplaceMixin, RetrieveMixin)


class CompanyService(CreateMixin, ListMixin, ModelMixin, ReplaceMixin, RetrieveMixin):
    api_url = "companies"

    def __init__(self, keap):
        super().__init__(keap)



from keap.XML import BaseService


class ContactService(BaseService):
    def __init__(self, keap):
        super().__init__(keap)

    def add(self, values):
        return self.call("add", values)

    def add_with_dupe_check(self, values, dup_check_type):
        return self.call("addWithDupCheck", values, dup_check_type)

    def update(self, contact_id, fields):
        return self.call("update", contact_id, fields)

    def load(self, contact_id, fields):
        return self.call("load", contact_id, fields)

    def delete(self, table, contact_id):
        return self.call("delete", table, contact_id)

    def merge(self, receiving_id, providing_id):
        return self.call("merge", receiving_id, providing_id)

    def find_by_email(self, email, fields):
        return self.call("findByEmail", email, fields)

    def add_to_group(self, contact_id, tag_id):
        return self.call("addToGroup", contact_id, tag_id)

    def add_tag(self, contact_id, tag_id):
        return self.add_to_group(contact_id, tag_id)

    def remove_from_group(self, contact_id, tag_id):
        return self.call("removeFromGroup", contact_id, tag_id)

    def remove_tag(self, contact_id, tag_id):
        return self.remove_from_group(contact_id, tag_id)

    def link_contacts(self, contact_id_1, contact_id_2, link_type_id):
        return self.call("linkContacts", contact_id_1, contact_id_2, link_type_id)

    def unlink_contacts(self, contact_id_1, contact_id_2, link_type_id):
        return self.call("unlinkContacts", contact_id_1, contact_id_2, link_type_id)

    def list_linked_contacts(self, contact_id):
        return self.call("listLinkedContacts", contact_id)

    def get_app_setting(self, module, setting):
        return self.call("getAppSetting", module, setting)



from keap.REST.V1.mixins import (CreateMixin, CreateCustomFieldMixin,  ListMixin, ModelMixin, ReplaceMixin,
                                 RetrieveMixin, UpdateMixin)


class OpportunityService(CreateMixin, CreateCustomFieldMixin, ListMixin, ModelMixin, ReplaceMixin, RetrieveMixin,
                         UpdateMixin):
    api_url = "opportunities"

    def __init__(self, keap):
        super().__init__(keap)



from keap.XML import BaseService


class SearchService(BaseService):
    def __init__(self, keap):
        super().__init__(keap)

    def get_search_report(self, savedSearchId, userId, pageNumber):
        return self.call("getSavedSearchResultsAllFields", savedSearchId, userId, pageNumber)


from keap.REST.V1.mixins import RetrieveMixin, DeleteMixin


class TagService(DeleteMixin, RetrieveMixin):
    api_url = "tags"
    return_key = "tags"

    def __init__(self, keap):
        super().__init__(keap)

    def list(self,
             category: int = None,
             name: str = None,  # This is a full string match, no wildcard
             limit: int = 1000,
             offset: int = 0,
             ):
        params = {
            'limit': limit,
            'offset': offset,
        }
        if category:
            params['category'] = category
        if name:
            params['name'] = name

        return self._get(f"{self.api_url}", params=params)

    def create(self, category: int, name: str, description: str = None):
        data = {
            'category': {
                'id': category
            },
            'name': name
        }
        if description:
            data['description'] = description
        return self._post(f"{self.api_url}", data=data)

    def create_category(self, name: str, description: str = None):
        data = {
            'name': name
        }
        if description:
            data['description'] = description
        return self._post(f"{self.api_url}/categories", data=data)

    def get_tagged_companies(self, id: int, limit: int = 1000, offset: int = 0):
        params = {
            'tagId': id,
            'limit': limit,
            'offset': offset,
        }
        return self._get(f"{self.api_url}/{id}/companies", params=params)

    def get_tagged_contacts(self, id: int, limit: int = 1000, offset: int = 0):
        params = {
            'tagId': id,
            'limit': limit,
            'offset': offset,
        }
        return self._get(f"{self.api_url}/{id}/contacts", params=params)

    def add_contacts(self, id: int, contact_ids: list):
        data = {
            'ids': contact_ids,
        }
        return self._post(f"{self.api_url}/{id}/contacts", data=data)

    def remove_contacts(self, id: int, contact_ids: list):
        params = {
            'ids': ','.join([str(cid) for cid in contact_ids]),
        }
        return self._delete(f"{self.api_url}/{id}/contacts", params=params)

    def remove_contact(self, id: int, contact_id: int):
        return self._delete(f"{self.api_url}/{id}/contacts/{contact_id}")



from keap.REST.V1.mixins import CreateMixin, ListMixin


class UserService(CreateMixin, ListMixin):
    api_url = "users"

    def __init__(self, keap):
        super().__init__(keap)

    def get_logged_in_user_info(self, **kwargs):
        return self._get(f"oauth/connect/userinfo", params=kwargs)

    def get_user_email_signature(self, user_id: int):
        return self._get(f"{self.api_url}/{user_id}/signature")



from keap.REST.V1.mixins import CreateMixin, ListMixin, ModelMixin


class SubscriptionService(CreateMixin, ListMixin, ModelMixin):
    api_url = "subscriptions"

    def __init__(self, keap):
        super().__init__(keap)



from keap.REST.V1.mixins import ListMixin, RetrieveMixin


class TransactionService(ListMixin, RetrieveMixin):
    api_url = "transactions"

    def __init__(self, keap):
        super().__init__(keap)



import json
from urllib import parse

import click as click
from click import prompt

from keap import Keap
from keap.settings import api_settings, IN_MEMORY_STORAGE, JSON_STORAGE


@click.group()
def cli():
    """
    Simple CLI for managing Keap API Access
    """
    pass


@cli.command()
def generate_client_config():
    client_id = prompt("What is your client id?", hide_input=True)
    client_secret = prompt("What is your client secret?", hide_input=True)
    redirect_url = prompt("Redirect URL", default=api_settings.REDIRECT_URL)
    app_name = prompt("App Name", default=api_settings.APP_NAME)
    allow_none = prompt("Allow None", default=api_settings.ALLOW_NONE)
    use_datetime = prompt("Use Datetime", default=api_settings.USE_DATETIME)
    storage_class = prompt("Storage Class", default=api_settings.defaults.get('STORAGE_CLASS'),
                           type=click.Choice(api_settings.defaults.get('DEFAULT_STORAGE_CLASSES')),
                           show_choices=True)

    creds = {
        'CLIENT_ID': client_id,
        'CLIENT_SECRET': client_secret,
        'REDIRECT_URL': redirect_url,
        'APP_NAME': app_name,
        'ALLOW_NONE': allow_none,
        'USE_DATETIME': use_datetime,
        'STORAGE_CLASS': storage_class,
    }

    if storage_class == JSON_STORAGE:
        creds['JSON_STORAGE_PATH'] = prompt("Token Storage File", default=api_settings.JSON_STORAGE_PATH,
                                            type=click.Path(readable=True, writable=True))

    save_to_file = prompt("Save settings to file?", default='y', type=click.BOOL, show_choices=True)

    if save_to_file:
        file = prompt("Save settings to file?", default=api_settings.CREDENTIALS_PATH, type=click.File("w", ))
        file.write(json.dumps(creds, indent=4))
    else:
        click.echo(creds)

    return creds


@cli.command()
@click.option('--credentials-file', '--f', type=click.File('r'), default=api_settings.CREDENTIALS_PATH,
              prompt="Path to Credentials File")
@click.pass_context
def get_access_token(ctx, credentials_file):
    """OAuth flow for Keap to obtain an access and refresh token"""
    try:
        creds = json.load(credentials_file)
    except Exception as e:
        return

    creds['APP_NAME'] = prompt("What app is being used?", type=click.STRING, default=creds['APP_NAME'])

    keap = Keap(
        config=creds
    )
    auth_url = keap.get_authorization_url()
    click.echo(f"Visit {auth_url}")
    response_url = prompt("Paste the return url here")
    query_string = dict(parse.parse_qsl(parse.urlsplit(response_url).query))
    code = query_string.get('code')
    if code:
        keap.request_access_token(code)

    if keap.token.access_token:
        if creds.get('STORAGE_CLASS') == IN_MEMORY_STORAGE:
            if prompt("You have chosen memory storage so nothing is saved here. Echo results?", default="y",
                      type=click.BOOL):
                click.echo(keap.token.__dict__)
            else:
                click.echo(f"Saved to {credentials_file}")


@cli.command()
@click.option('--credentials-file', '--cf', type=click.File("r"), default=api_settings.CREDENTIALS_PATH)
@click.option('--app', '--a', type=click.STRING, default='default')
def refresh_access_token(credentials_file, app):
    """Refresh Keap access token"""
    try:
        creds = json.load(credentials_file)
    except Exception as e:
        click.echo("Failed to load credentials file")
        return

    if app:
        creds['APP_NAME'] = app

    keap = Keap(
        config=creds
    )
    keap.refresh_access_token()


if __name__ == "__main__":
    cli()



from setuptools import setup, find_packages

setup(
    name='keap-python',
    version='1.1.2',
    description='Python SDK for Keap API with Django Integration',
    url='https://bitbucket.org/theapiguys/keap-python',
    author='Brandon @ TheAPIGuys',
    author_email='brandon@theapiguys.com',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Click',
        'requests'
    ],
    entry_points={
        'console_scripts': [
            'keap = keap.scripts.cli:cli',
        ],
    },
)



from keap.REST.V1.mixins import (CreateMixin, CreateCustomFieldMixin, DeleteMixin, ListMixin, ModelMixin, ReplaceMixin,
                                 RetrieveMixin, UpdateMixin)


class NoteService(CreateMixin, CreateCustomFieldMixin, DeleteMixin, ListMixin, ModelMixin, ReplaceMixin,
                  RetrieveMixin, UpdateMixin):
    api_url = "notes"

    def __init__(self, keap):
        super().__init__(keap)

    def get_contact_notes(self,
                          contact_id: int = None):
        params = locals()
        return self._get(f"{self.api_url}", params=params)
